# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides graphs for FigureOfMerit objects."""
import collections.abc
import typing

import matplotlib

from ansys.stk.core.stkobjects import FigureOfMerit
from ansys.stk.extensions.data_analysis.graphs._common_graphs import (
    _grid_stats_over_time_line_chart,
    _value_by_latitude_xy_graph,
    _value_by_longitude_xy_graph,
)
from ansys.stk.extensions.data_analysis.graphs.graph_helpers import interval_plot, line_chart


def gi_all_dop_line_chart(stk_object :FigureOfMerit, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of all DOP values, over time, for the point currently selected via the figure of merit grid inspector.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\FigureOfMerit\\GI All DOP.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.FigureOfMerit
        The STK FigureOfMerit object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Selected Point All DOPs').execute_elements(start_time, stop_time, step, ['VDOP', 'Time', 'TDOP', 'PDOP', 'GDOP', 'EDOP', 'HDOP', 'NDOP']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : False, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Value', 'lines': [
            {'y_name':'gdop', 'label':'GDOP', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'},
            {'y_name':'pdop', 'label':'PDOP', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'},
            {'y_name':'hdop', 'label':'HDOP', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'},
            {'y_name':'vdop', 'label':'VDOP', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'},
            {'y_name':'edop', 'label':'EDOP', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'},
            {'y_name':'ndop', 'label':'NDOP', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'},
            {'y_name':'tdop', 'label':'TDOP', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'}]}]
    return line_chart([df], root, ['gdop','pdop','hdop','vdop','edop','ndop','tdop'], ['time'], axes, 'time', 'Time', 'All Dilutions of Precision', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def gi_point_fom_line_chart(stk_object :FigureOfMerit, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the figure of merit values over time, for the point currently selected via the figure of merit grid inspector.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\FigureOfMerit\\GI Point FOM.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.FigureOfMerit
        The STK FigureOfMerit object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Selected Point FOM').execute_elements(start_time, stop_time, step, ['Time', 'FOM Value']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : False, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Value', 'lines': [
            {'y_name':'fom value', 'label':'FOM Value', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'}]}]
    return line_chart([df], root, ['fom value'], ['time'], axes, 'time', 'Time', 'Point FOM Value', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def gi_point_satisfaction_interval_graph(stk_object :FigureOfMerit, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create an interval graph of the satisfaction intervals for the point currently selected via the figure of merit grid inspector.

    Satisfaction intervals are defined as periods when a grid point achieves the defined satisfaction criteria associated with the FOM.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\FigureOfMerit\\GI Point Satisfaction.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.FigureOfMerit
        The STK FigureOfMerit object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Selected Point Satisfaction').execute_elements(start_time, stop_time, ['Interval Start', 'Interval End']).data_sets.to_pandas_dataframe()
    elements=[(('interval start', 'None'),('interval end', 'None'))]
    return interval_plot([df], root, elements, [], ['interval start','interval end'], 'Time', 'Point Satisfaction Intervals', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def gi_region_fom_line_chart(stk_object :FigureOfMerit, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the minimum, maximum, and average figure of merit value, sampled over all grid points within the region currently selected in the figure of merit grid inspector, over time.

    Grid points for which a value cannot be computed are not included in the reported statistics.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\FigureOfMerit\\GI Region FOM.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.FigureOfMerit
        The STK FigureOfMerit object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Selected Region FOM').execute_elements(start_time, stop_time, step, ['Time', 'Minimum', 'Maximum', 'Average']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : False, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Value', 'lines': [
            {'y_name':'minimum', 'label':'Minimum', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'},
            {'y_name':'maximum', 'label':'Maximum', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'},
            {'y_name':'average', 'label':'Average', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'}]}]
    return line_chart([df], root, ['minimum','maximum','average'], ['time'], axes, 'time', 'Time', 'Regional FOM Values', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def gi_region_satisfaction_interval_graph(stk_object :FigureOfMerit, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create an interval graph of the intervals of time when the region selected by the grid inspector is partially covered.

    A region is considered to be covered if at least one point within the region has access to one or more of the assigned assets.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\FigureOfMerit\\GI Region Satisfaction.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.FigureOfMerit
        The STK FigureOfMerit object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Selected Region Partial Satisfaction').execute_elements(start_time, stop_time, ['Interval Start', 'Interval End']).data_sets.to_pandas_dataframe()
    elements=[(('interval start', 'None'),('interval end', 'None'))]
    return interval_plot([df], root, elements, [], ['interval start','interval end'], 'Time', 'Periods of Partial Regional Satisfaction', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def grid_stats_over_time_line_chart(stk_object :FigureOfMerit, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the minimum, maximum, and average figure of merit values, sampled over all grid points, over time.

    Grid points for which a value cannot be computed are not included in the reported statistics.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\FigureOfMerit\\Grid Stats Over Time.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.FigureOfMerit
        The STK FigureOfMerit object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _grid_stats_over_time_line_chart(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

def satisfied_by_time_line_chart(stk_object :FigureOfMerit, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the percentage of the grid which satisfies the satisfaction criteria defined in the figure of merit, as a function of time.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\FigureOfMerit\\Satisfied By Time.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.FigureOfMerit
        The STK FigureOfMerit object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Satisfied by Time').execute_elements(start_time, stop_time, step, ['Time', 'Percent Satisfied']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : False, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Percent', 'lines': [
            {'y_name':'percent satisfied', 'label':'% Satisfied', 'use_unit':False, 'unit_squared': None, 'dimension': 'Percent'}]}]
    return line_chart([df], root, ['percent satisfied'], ['time'], axes, 'time', 'Time', 'Satisfied By Time', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def value_by_latitude_line_chart(stk_object : FigureOfMerit, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the minimum, maximum, and average figure of merit value, sampled over all grid points at the same latitude, as a function of latitude.

    Statistics are generated by sampling values from all grid points with latitude values within one half degree of the reported latitude value. Values are computed for every one degree of latitude. For example, statistics reported for the latitude value of 30 degrees will represent figure of merit values for all points with latitudes between 29.5 and 30.5 degrees. Grid points for which a value cannot be computed are not included in the reported statistics. Latitudes which do not have any reported grid points within one half degree are not included in the reported values.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\FigureOfMerit\\Value By Latitude.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.FigureOfMerit
        The STK FigureOfMerit object.
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _value_by_latitude_xy_graph(stk_object, colormap, time_unit_abbreviation, formatter)

def value_by_longitude_line_chart(stk_object : FigureOfMerit, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the minimum, maximum, and average figure of merit value, sampled over all grid points at the same longitude, as a function of longitude.

    Statistics are generated by sampling values from all grid points with longitude values within one half degree of the reported longitude value. Values are computed for every one degree of longitude. For example, statistics reported for the longitude value of 30 degrees will represent figure of merit values for all points with longitudes between 29.5 and 30.5 degrees. Grid points for which a value cannot be computed are not included in the reported statistics. Longitudes which do not have any reported grid points within one half degree are not included in the reported values.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\FigureOfMerit\\Value By Longitude.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.FigureOfMerit
        The STK FigureOfMerit object.
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _value_by_longitude_xy_graph(stk_object, colormap, time_unit_abbreviation, formatter)




# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""PySTK Graphing Utilities.

A set of helper functions for graphing basic STK desktop graph types.
"""

import collections.abc
from math import ceil, radians
import typing

import matplotlib
import matplotlib.axes
import matplotlib.figure
import matplotlib.pyplot
import numpy as np
import pandas

from ansys.stk.core.stkobjects import Access, STKObjectRoot
from ansys.stk.core.stkutil import UnitPreferencesDimensionCollection
from ansys.stk.extensions.data_analysis._dates import _STKDate, _STKDateConverter


def polar_chart(data : list[pandas.DataFrame], root : STKObjectRoot, numerical_columns : list[str], axis : dict, title : str, origin_0 : bool = False, convert_negative_r : bool = False, colormap: matplotlib.colors.Colormap = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Create a polar chart from the provided dataframe and axis information.

    Parameters
    ----------
    data : list of pandas.DataFrame
        The list of DataFrames containing the data.
    root : ansys.stk.core.stkobjects.STKObjectRoot
        The STK object root.
    numerical_columns : list of str
        The list of dataframe columns with numerical values.
    axis : dict
        The dictionary containing information about the data to plot.
    title : str
        The title of the chart.
    origin_0 : bool
        Whether to set the theta 0 point to the top of the graph.
    convert_negative_r : bool
        Whether to convert negative radius values by using opposite angle values.
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the lines (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    # create plot
    fig, ax = matplotlib.pyplot.subplots(subplot_kw={"projection": "polar"})

    # get unit preferences from root
    units_preferences = root.units_preferences
    unit = None

    # use one color per line
    colors = colormap(np.linspace(0, 1, len(data))) if colormap else matplotlib.pyplot.cm.rainbow(np.linspace(0,1,len(data)))

    for i, df in enumerate(data):
        # data conversions
        df = _convert_columns(df, numerical_columns, [], units_preferences)
        df.dropna(axis=0, inplace=True)

        # matplotlib works in radians, so get x (theta) column and convert
        x_column = axis["lines"][0]["x_name"]
        df[x_column] = df[x_column].apply(lambda x : radians(x))

        # get line information
        line = axis["lines"][0]
        # get y (r) variable
        y_var = line["y_name"]

        # matplotlib doesn"t support negative r values in polar graphs, so convert for stk graphs
        # that show negative r values with negative angle
        if convert_negative_r:
            _eliminate_negative_r_polar_vals(df, y_var, x_column)

        # if line uses unit, get current unit
        label = ""
        if line["use_unit"]:
            unit = units_preferences.get_current_unit_abbrv(line["dimension"])
            label = line["label"] + f"({unit})"

        # plot x and y data
        x_data = df[x_column]
        y_data = df[y_var]
        ax.plot(x_data, y_data, label=label, color=colors[i])

    # set x label
    ax.set_xlabel(f"{axis['label']} ({unit})") if unit else ax.set_xlabel(f"{axis['label']}")
    # set styling
    ax.set_facecolor("whitesmoke")
    ax.grid(visible=True, axis="both", which="both", linestyle="--")
    # set title
    ax.set_title(title, y=1.05)
    # set theta direction to match stk
    ax.set_theta_direction(-1)

    # styling for y axis with negative values to match stk styling
    if not convert_negative_r:
        ax.invert_yaxis()

    # configure origin location to match stk
    if origin_0:
        ax.set_theta_zero_location("N")

    # return figure and axis
    return fig, ax

def interval_plot(data : list[pandas.DataFrame], root : STKObjectRoot, element_pairs : list, numerical_columns : list[str], time_columns : list[str], x_label : str, title : str, colormap: matplotlib.colors.Colormap = None, time_unit_abbreviation: str = "UTCG", formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Create an interval plot from the provided list of dataframes.

    Parameters
    ----------
    data : list of pandas.DataFrame
        The list of DataFrames containing the data.
    root : ansys.stk.core.stkobjects.STKObjectRoot
        The STK object root.
    numerical_columns : list of str
        The list of dataframe columns with numerical values.
    time_columns : list of str
        The list of dataframe columns with time values.
    x_label : str
        The label for the x-axis.
    title : str
        The title of the chart.
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the lines (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    # create plot
    fig, ax = matplotlib.pyplot.subplots()

    # count number of bars
    bar_num = 0

    # create color map with one color for each bar
    colors = colormap(np.linspace(0, 1, len(element_pairs))) if colormap else matplotlib.pyplot.cm.rainbow(np.linspace(0, 1, len(element_pairs)))

    # collect bars
    bars = []

    # collect y tick locations
    tick_locs = []

    # get unit preferences from root
    units_preferences = root.units_preferences

     # data conversions
    for df in data:
        df = _convert_columns(df, numerical_columns, time_columns, units_preferences, root=root)
        df.dropna(axis=0, inplace=True)

    # format time axis
    all_times = (pandas.concat([df[time_columns] for df in data])).stack().reset_index(drop=True).sort_values()
    time_difference = all_times.iloc[-1] - all_times.iloc[0]
    matplotlib.units.registry[_STKDate] = _STKDateConverter(root, time_difference, time_unit_abbreviation, formatter)

    # iterate through pairs of elements
    for i in range(len(element_pairs)):
        element_pair = element_pairs[i]
        first_elem = element_pair[0]
        second_elem = element_pair[1]
        # get corresponding dataframe
        df = data[i]
        # add column to dataframe with difference between end and start times
        df["graph duration"] = df[second_elem[0]] - df[first_elem[0]]
        # create bar starting at start times with length corresponding to duration
        # label if label provided, otherwise leave blank
        bars.append(ax.broken_barh(list(zip(df[first_elem[0]], df["graph duration"])), (bar_num*10 + 10, 9), zorder=1, facecolors=colors[bar_num], label=first_elem[1] if first_elem[1] else ""))
        # append tick location
        tick_locs.append(bar_num*10 + 14.5)
        bar_num += 1

    # set x label
    ax.set_xlabel(f"{x_label}")

    # create legend if more than one bar
    if len(element_pairs) > 1:
        ax.legend(bars, [b.get_label() for b in bars])

    # set title
    ax.set_title(title)

    # set styling
    ax.set_facecolor("whitesmoke")
    ax.grid(visible=True, axis="both", which="both", linestyle="--")
    ax.set_axisbelow(True)

    # label y-axis using tick locations
    ax.set_yticks(tick_locs, labels = [b.get_label() for b in bars])

    # hide y-axis
    ax.get_yaxis().set_visible(False)

    # set size
    fig.set_size_inches(18.5, 7)

    _format_time_x_axis(fig, ax, all_times.iloc[0], all_times.iloc[-1], root, time_unit_abbreviation)

    # return figure and axis
    return fig, ax

def line_chart(data : list[pandas.DataFrame], root : STKObjectRoot, numerical_columns : list[str], time_columns: list[str], axes : list[dict], x_column : str, x_label : str, title : str, colormap: matplotlib.colors.Colormap = None, time_unit_abbreviation: str = "UTCG", formatter: collections.abc.Callable[[float, float], str] = None, multiple_data_providers: bool = False) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Create a line chart from the provided dataframe and axes information.

    Parameters
    ----------
    data : list of pandas.DataFrame
        The list of DataFrames containing the data.
    root : ansys.stk.core.stkobjects.STKObjectRoot
        The STK object root.
    numerical_columns : list of str
        The list of dataframe columns with numerical values.
    time_columns : list of str
        The list of dataframe columns with time values.
    axes : list of dict
        The list of dictionaries containing information about the data to plot.
    x_column : str
        The column corresponding to the x-axis data.
    x_label : str
        The label for the x-axis.
    title : str
        The title of the chart.
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the lines (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).
    multiple_data_providers: bool
        Whether each dataframe provided corresponds to a different data provider and axis (the default is False).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    # count number of lines
    num_lines = 0
    for axis in axes:
        num_lines += len(axis["lines"])

    # get unit preferences from root
    units_preferences = root.units_preferences

    # create plot
    fig, ax = matplotlib.pyplot.subplots()
    # add matplotlib axis to list
    axes_list = [ax]

    # data conversions
    for i, df in enumerate(data):
        df.dropna(inplace=True, subset=[x_column])
        if multiple_data_providers:
            numerical_columns_to_convert = numerical_columns[i]
        else:
            numerical_columns_to_convert = numerical_columns
        if time_columns:
            df = _convert_columns(df, numerical_columns_to_convert, time_columns, units_preferences, root=root)
        elif numerical_columns:
            df = _convert_columns(df, numerical_columns_to_convert, time_columns, units_preferences)
        df.sort_values(x_column, inplace=True)
    if x_column in time_columns:
        all_times = (pandas.concat([df[x_column] for df in data])).sort_values()
        time_difference = all_times.iloc[-1] - all_times.iloc[0]
        matplotlib.units.registry[_STKDate] = _STKDateConverter(root, time_difference, time_unit_abbreviation, formatter)

    # used to count line number to subset color map
    line_count = 0
    # line collection for legend
    mpl_lines = []
    for i in range(len(data)):
        df = data[i]

        # create color map with correct length
        colors = colormap(np.linspace(0, 1, num_lines)) if colormap else matplotlib.pyplot.cm.rainbow(np.linspace(0, 1, num_lines))

        # used to count line number to subset color map
        if not multiple_data_providers:
            line_count = 0
            # line collection for legend
            mpl_lines = []

        # get x data from dataframe
        x_data = df[x_column]

        # iterate through axes information parameter
        for j in range(len(axes)):
            axis = axes[j]
            # if additional axes needed, duplicated matplotlib axis
            if j != 0 and i==0:
                ax = ax.twinx()
                axes_list.append(ax)
            ax = axes_list[j]
            # iterate through lines under axis
            for k in range(len(axis["lines"])):
                if (not multiple_data_providers) or (multiple_data_providers and i == j):
                    line = axis["lines"][k]
                    # get y data
                    y_data = df[line["y_name"]]
                    # get line label
                    label = line["label"]
                    # if line uses unit, get current unit
                    if line["use_unit"]:
                        unit = units_preferences.get_current_unit_abbrv(line["dimension"])
                        # check if unit should be squared in label
                        if line["unit_squared"]:
                            label = label + f"({unit}^2)"
                        else:
                            label = label + f"({unit})"

                    mpl_lines.extend(ax.plot(x_data, y_data, label=label, color=colors[line_count], linewidth=1.5))
                    line_count += 1

            if i == 0:
                # if axis uses unit, set unit in label
                if axis["use_unit"]:
                    if axis["unit_squared"]:
                        ax.set_ylabel(axis["label"] + f" ({unit}^2)")
                    else:
                        ax.set_ylabel(axis["label"] + f" ({unit})")
                else:
                    ax.set_ylabel(axis["label"])

                # set x-label
                ax.set_xlabel(x_label)

                # set styling (must be done for each axis)
                ax.set_facecolor("whitesmoke")

                if len(axes) == 1:
                    ax.grid(visible=True, axis="both", which="both", linestyle="--")
                # if multiple axes, only plot x gridlines
                else:
                    ax.grid(visible=True, axis="x", which="both", linestyle="--")

                # set axis scales
                if axis["ylog10"]:
                    ax.set_yscale("log")
                elif axis["y2log10"]:
                    ax.set_yscale("log", base=2)

    # format time x-axis if needed
    if x_column in time_columns:
        _format_time_x_axis(fig, ax, all_times.iloc[0], all_times.iloc[-1], root, time_unit_abbreviation)

    # if multiple lines, create legend
    if num_lines > 1:
        legend = ax.legend(mpl_lines, [line.get_label() for line in mpl_lines], loc="upper center", fontsize=10)
        for legend_object in legend.legend_handles:
            legend_object.set_linewidth(2.0)

    # set title and size
    ax.set_title(title)
    fig.set_size_inches(12, 6)
    # return figure and axis
    return fig, ax

def pie_chart(
    root: STKObjectRoot,
    df: pandas.DataFrame,
    numerical_columns: list[str],
    time_columns: list[str],
    column: str,
    title: str,
    dimension: str,
    label_column: str = None,
    colormap: matplotlib.colors.Colormap = None
) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Create a pie chart from the provided dataframe and information.

    Parameters
    ----------
    root : ansys.stk.core.stkobjects.STKObjectRoot
        The STK object root.
    df : pandas.DataFrame
        The dataframe containing the data.
    numerical_columns : list of str
        The list of dataframe columns with numerical values.
    time_columns : list of str
        The list of dataframe columns with time values.
    column : str
        The dataframe column to plot in the pie chart.
    title : str
        The title of the chart.
    dimension : str
        The dimension of the chart data.
    label_column : str
        The dataframe column corresponding to the chart labels (the default is None).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the pie chart slices (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.

    """
    # create plot
    fig, ax = matplotlib.pyplot.subplots()

    # get units
    units_preferences = root.units_preferences
    unit = units_preferences.get_current_unit_abbrv(dimension)

    # data conversions
    df = _convert_columns(df, numerical_columns, time_columns, units_preferences, root= root)
    df.dropna(axis=0, inplace=True)

    # create colormap with one color for each slice of pie
    num_colors_needed = len(df[column].value_counts())
    colors = colormap(np.linspace(0, 1, num_colors_needed)) if colormap else matplotlib.pyplot.cm.rainbow(np.linspace(0, 1, num_colors_needed))

    # if plot is labeled with a different column, get and configure labels
    labels = []
    if label_column:
        for i in range(len(df[label_column])):
            labels.append(f"{label_column} {df[label_column][i]:.0f}: {df[column][i]:.3f}({unit})")

    # create pie chart
    ax.pie(df[column], autopct="%1.1f%%", labels=labels, colors=colors, textprops={"fontsize": 8}, counterclock=False)

    # set title
    ax.set_title(title)

    # return figure and axis
    return fig, ax


def interval_pie_chart(
    root: STKObjectRoot,
    df: pandas.DataFrame,
    numerical_columns: list[str],
    time_columns: list[str],
    start_column: str,
    stop_column: str,
    start_time: str,
    stop_time: str,
    title: str,
    dimension: str,
    cumulative: bool = False,
    color_list: list[typing.Any] = None
) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Create an interval pie chart from the provided dataframe.

    Parameters
    ----------
    root : ansys.stk.core.stkobjects.STKObjectRoot
        The STK object root.
    df : pandas.DataFrame
        The dataframe containing the data.
    numerical_columns : list of str
        The list of dataframe columns with numerical values.
    time_columns : list of str
        The list of dataframe columns with time values.
    start_column : str
        The column containing the start times of the intervals.
    stop_column : str
        The column containing the stop times of the intervals.
    start_time : str
        The start time of the calculation.
    stop_time : str
        The stop time of the calculation.
    title : str
        The title of the chart.
    dimension : str
        The dimension of the chart data.
    cumulative : bool
        Whether the intervals should be summed into durations and gaps (the default is False).
    color_list : list of typing.Any
        The colors with which to color the pie chart slices (the default is None). Must have length >= 2.

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.

    Raises
    ------
    ValueError
        If the length of color_list is less than 2.
    """
    # get unit preferences
    units_preferences = root.units_preferences
    unit = units_preferences.get_current_unit_abbrv(dimension)
    date_unit = units_preferences.get_current_unit_abbrv("Date")

    # data conversions
    df = _convert_columns(df, numerical_columns, time_columns, units_preferences, root=root)

    # create duration column using stop and start columns
    df["graph duration"] = df[stop_column] - df[start_column]

    # create gap column with times in between durations
    df["graph gap"] = df[start_column].shift(-1) - df[stop_column]

    # last gap is from end of last duration to stop time
    last_gap = _STKDate.from_value_and_format(root, stop_time, unit=date_unit) - df.iloc[-1][stop_column]
    #if last_gap != 0:
    df.at[df.index[-1], "graph gap"] = last_gap
    # first gap is from analysis start time to start of first duration
    first_gap = df[start_column][0] - _STKDate.from_value_and_format(root, start_time, unit=date_unit)

    # create plot
    fig, ax = matplotlib.pyplot.subplots()
    if color_list and len(color_list) < 2:
        raise ValueError("If provided, 'color_list' argument must contain at least 2 colors.")

    if cumulative:
        # if plot is cumulative, sum durations
        duration_sum = df["graph duration"].sum()
        # then gap is equivalent to sum of gaps + first gap
        gap_sum = df["graph gap"].sum() + first_gap

        colors = color_list if color_list else ["deepskyblue", "slategray"]

        # plot duration and gap sums
        matplotlib.pyplot.pie(
            [duration_sum, gap_sum],
            labels=[
                f"Cumulative Duration: {duration_sum:.2f} ({unit})",
                f"Cumulative Gap: {gap_sum:.2f} ({unit})",
            ],
            colors=colors,
            autopct="%1.3f%%",
            labeldistance=None,
            pctdistance=1.2,
        )
        # create legend
        ax.legend(shadow=True, loc="lower center")
    else:
        colors = color_list if color_list else ["slategray", "deepskyblue"]

        # create zipped list with each duration and gap paired
        zip_list = list(zip(df["graph duration"], df["graph gap"]))
        flat_list = []
        label_list = []
        count = 2

        # flatten list while maintaining order and create list of labels
        for duration, gap in zip_list:
            flat_list.extend([duration, gap])

            if not np.isnan(duration) and duration != 0:
                label_list.append(f"duration {count -1}: {duration:.2f}({unit})")

            if not np.isnan(gap) and gap !=0:
                label_list.append(f"gap {count}: {gap:.2f}({unit})")

            count += 1

        # remove any nan or 0 values
        cleaned_list = [x for x in flat_list if not np.isnan(x) and x!=0]
        # get gap before start of first interval, add to data and label lists
        if first_gap != 0:
            cleaned_list.insert(0, first_gap)
            label_list.insert(0, f"gap 1: {first_gap:.2f}({unit})")
        # if no first gap, reverse color order to maintain blue for durations and gray for gaps
        else:
            colors.reverse()
        # plot intervals
        matplotlib.pyplot.pie(
            cleaned_list,
            colors=colors,
            autopct="%1.3f%%",
            pctdistance=1.15,
            wedgeprops={"edgecolor": "black", "linewidth": 1, "antialiased": True},
        )
        # set size
        fig.set_size_inches(8, 8)
        # create legend
        ax.legend(
            labels=label_list,
            ncol=ceil(len(cleaned_list) / 4),
            loc="lower center",
            bbox_to_anchor=(0.5, -0.15),
            shadow=True,
            fontsize="x-small",
        )

    # set title
    ax.set_title(title)

    # return figure and axis
    return fig, ax

def _convert_columns(
    df: pandas.DataFrame, numerical_column_list: list[str], date_column_list: list[str], unit_preferences: UnitPreferencesDimensionCollection, root: STKObjectRoot = None
) -> pandas.DataFrame:
    """Convert numerical and time columns in a pandas dataframe.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        The dataframe containing the data.
    numerical_column_list : list of str
        The list of dataframe columns with numerical values.
    date_column_list : list of str
        The list of dataframe columns with time values.
    unit_preferences : UnitPreferencesDimensionCollection
        The collection of scenario-wide unit preferences.
    root : STKObjectRoot
        The STK object root.

    Returns
    -------
    pandas.DataFrame
        The dataframe with converted columns.
    """
    df[numerical_column_list] = df[numerical_column_list].astype(float)
    if date_column_list:
        for col in date_column_list:
            if col in df:
                df[col] = df[col].apply(lambda x: _STKDate.from_value_and_format(root, x, unit_preferences.get_current_unit_abbrv("Date")))
    return df

def _format_time_x_axis(fig : matplotlib.figure.Figure, ax : matplotlib.axes.Axes, first_time : _STKDate, last_time : _STKDate, root : STKObjectRoot, time_unit_abbreviation : str):
    """Convert numerical and time columns in a pandas dataframe.

    Parameters
    ----------
    matplotlib.figure.Figure
        The figure.
    matplotlib.axes.Axes
        The axes.
    first_time : _STKDate
        The earliest time in the data.
    last_time : _STKDate
        The latest time in the data.
    root : STKObjectRoot
        The STK object root.
    """

    def get_d_m_y_utcg(date : _STKDate):
        return date.get_utcg().rsplit(" ", maxsplit=1)[0]

    if time_unit_abbreviation == "UTCG":
        if last_time - first_time < 604800:
            fig.text(0.05, 0.01, get_d_m_y_utcg(first_time),
                horizontalalignment="left", verticalalignment="bottom",
                fontsize=10)
            if get_d_m_y_utcg(first_time) != get_d_m_y_utcg(last_time):
                fig.text(0.95, 0.01, get_d_m_y_utcg(last_time),
                horizontalalignment="right", verticalalignment="bottom",
                fontsize=10)

    # add vertical lines showing day changes
    if last_time - first_time < 2592000:
        start_date = _STKDate.from_value_and_format(root, get_d_m_y_utcg(first_time) + " 00:00:00.000").add_duration(1, "day")
        while start_date < last_time:
            ax.axvline(x=start_date.get_epsec(), color="black", linestyle="-", linewidth=1.5)
            ax.annotate(start_date.format(time_unit_abbreviation), xy =(start_date.get_epsec(),ax.get_ylim()[0]), xytext=(0, 7), rotation = 270, fontsize=9, textcoords="offset points")
            start_date = start_date.add_duration(1, "day")

def _get_access_data(access :Access, item : str, group : bool, group_name : str, elements: list[str], start_time: typing.Any, stop_time: typing.Any, step : float) -> list[pandas.DataFrame]:
    """Get list of data for access object, grouping by access interval while respecting start and stop times.

    Parameters
    ----------
    access : ansys.stk.core.stkobjects.Access
        The STK Access object.
    item : str
        The data provider.
    group : bool
        If the data provider is grouped.
    group_name : str
        The group.
    elements : list of str
        The list of data provider elements to execute.
    start_time : typing.Any
        The start time of the calculation.
    stop_time : typing.Any
        The stop time of the calculation.
    step_time : float
        The step time for the calculation.

    Returns
    -------
    list of pandas.DataFrame
        The list of data.

    Raises
    ------
    ValueError
        If none of the access intervals are contained within the provided start and stop times.
    """
    data=[]
    root = access.base.root
    start_time = _STKDate.from_value_and_format(root, start_time)
    stop_time = _STKDate.from_value_and_format(root, stop_time)
    access_intervals = access.computed_access_interval_times
    for i in range(0, access_intervals.count):
        times = access_intervals.get_interval(i)
        interval_start = _STKDate.from_value_and_format(root, times[0])
        interval_end = _STKDate.from_value_and_format(root, times[1])
        computation_start = None
        computation_stop = None
        # interval fully outside of desired calculation period, so skip
        if (interval_start < start_time and interval_end < start_time) or (interval_start > stop_time and interval_end > stop_time):
            continue
        # interval fully contained within desired calculation period, so include entire interval
        if interval_start >= start_time and interval_end <= stop_time:
            computation_start = interval_start.get_utcg()
            computation_stop = interval_end.get_utcg()
        else:
            # starts before desired calculation period, so start calculation at desired start time
            if interval_start < start_time:
                computation_start = start_time.get_utcg()
            # ends after desired calculation period, so end calculation at desired end time
            if interval_end > stop_time:
                computation_stop = stop_time.get_utcg()
        if group:
            data.append(access.data_providers.item(item).group.item(group_name).execute_elements(computation_start, computation_stop, step, elements).data_sets.to_pandas_dataframe())
        else:
            data.append(access.data_providers.item(item).execute_elements(computation_start, computation_stop, step, elements).data_sets.to_pandas_dataframe())
    if len(data) == 0:
        raise ValueError("No access data to plot- check provided start and stop times.")
    return data

def _eliminate_negative_r_polar_vals(df : pandas.DataFrame, r_var : str, theta_var : str):
    """Convert negative r values in a dataframe that has r and theta values.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe containing the data.
    r_var : str
        The column corresponding to the radius variable.
    theta_var : str
        The column corresponding to the angle variable.
    """
    df[theta_var]= np.where(df[r_var] >= 0, df[theta_var], df[theta_var] + np.pi)
    df[r_var] = df[r_var].apply(lambda x: abs(x))



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Records calls to an API."""

import inspect
import logging
from pathlib import Path
import re
import sys

from .recording import Recording


class Recorder:
    """
    Records the calls of interest.

    Uses :func:`sys.setprofile` to intercept calls.
    """

    def __init__(
        self, program, entry_point, root_directory, is_member_name_of_interest_func, run_as_module, program_args=None
    ):
        """Construct a new recorder."""
        self.program = program
        self.entry_point = entry_point
        self.root_directory_path = Path(root_directory).resolve()
        self.root_directory = root_directory
        self.is_member_name_of_interest_func = is_member_name_of_interest_func
        self.recording = Recording(self.root_directory_path)
        self.run_as_module = run_as_module
        self.program_args = program_args

    def record(self):
        """Record the execution of the specified program."""
        program_args_str = " ".join(self.program_args) if self.program_args else ""
        if not self.run_as_module:
            logging.info(f"Recording {self.program} {program_args_str}")
            script_filepath = Path(self.program)
            script_dirpath = str(script_filepath.parent.resolve())
            sys_path_append_cmd = f"sys.path.append('{script_dirpath}')".replace("\\", "\\\\\\\\")
            module_to_import = script_filepath.stem
        else:
            logging.info(f"Recording -m {self.program} {program_args_str}")
            sys_path_append_cmd = ""
            module_to_import = self.program

        # Temporarily modify sys.argv to pass arguments to the recordee
        prev_sys_argv = sys.argv
        new_sys_argv = [module_to_import]
        if self.program_args is not None:
            new_sys_argv += self.program_args
        sys.argv = new_sys_argv

        if sys.version_info >= (3, 11):
            exec_entry_point = "\n".join([
                f"import {module_to_import} as this_module",
                f"this_module.{self.entry_point}()"
            ])
        else:
            exec_entry_point = "\n".join([
                f"import {module_to_import}",
                f"{module_to_import}.{self.entry_point}()"
            ])

        bootstrap = [
            "exec('import sys')",
            f'exec("{sys_path_append_cmd}")',
            f'exec("""{exec_entry_point}""")',
            f"exec('del sys.modules[\"{module_to_import}\"]')",
        ]

        sys.setprofile(self._trace_call)

        [eval(cmd) for cmd in bootstrap] # nosec: B307

        sys.setprofile(None)

        # Restore initial arguments
        sys.argv = prev_sys_argv

        return self.recording

    def _get_client_frame_of_interest(self, frame):
        """Find the client frame calling the API of interest."""
        client_frame = frame.f_back
        client_filename = client_frame.f_code.co_filename
        client_filename_path = Path(client_filename).resolve()
        is_under_root_directory = (
            self.root_directory_path in client_filename_path.parents
        ) and client_filename_path.exists()
        logging.debug(
            f"Checking frame: name={frame.f_code.co_name}, root directory={self.root_directory_path}, client filename={Path(client_filename).resolve()}"
        )
        if is_under_root_directory:
            return client_frame

        lookup_frame = client_frame
        while lookup_frame is not None:
            if lookup_frame.f_code.co_name == "__setattr__":
                break
            lookup_frame = lookup_frame.f_back

        if (
            lookup_frame
            and "self" in frame.f_locals
            and "self" in lookup_frame.f_locals
            and id(frame.f_locals["self"]) == id(lookup_frame.f_locals["self"])
        ):
            client_frame = lookup_frame.f_back
            arg_info = inspect.getargvalues(lookup_frame)
            if len(arg_info.args) > 1 and arg_info.locals[arg_info.args[1]] == frame.f_code.co_name:
                client_filename = client_frame.f_code.co_filename
                client_filename_path = Path(client_filename).resolve()
                is_under_root_directory = (
                    self.root_directory_path in client_filename_path.parents
                ) and client_filename_path.exists()
                if is_under_root_directory:
                    return client_frame

        return None

    def _get_qualified_name(self, frame):
        """Get the qualified name for the function in the specified frame."""
        if hasattr(frame.f_code, "co_qualname"):
            # co_qualname was added in Python 3.11
            return frame.f_code.co_qualname
        else:
            # for previous versions, find the function object
            # in the garbage collector and get the qualified
            # name from it
            import gc
            import inspect

            for o in gc.get_objects():
                if inspect.isfunction(o) and id(o.__code__) == id(frame.f_code):
                    return o.__qualname__
        return None

    def _get_classname(self, frame, member):
        """Get the class name of the specified member in the specified frame."""
        if "self" in frame.f_locals:
            class_type = type(frame.f_locals["self"])
            if class_type is not None:
                types = list(class_type.__mro__)
                for t in types:
                    if member in t.__dict__:
                        return t.__name__
        else:
            qualified_name = self._get_qualified_name(frame)
            match = re.search("(?:.*\\.)?(.*)\\..*", qualified_name)
            if match:
                return match.group(1)
        return None

    def _trace_call(self, frame, event, arg):
        """Call back from :func:`sys.setprofile` to intercept all calls."""
        if event != "call":
            return

        is_call_of_interest = self.is_member_name_of_interest_func(frame.f_code.co_name)

        if not is_call_of_interest:
            return

        client_frame = self._get_client_frame_of_interest(frame)
        if client_frame:

            member = frame.f_code.co_name

            type_name = self._get_classname(frame, member)

            if type_name is not None:

                import inspect

                frameinfo = inspect.getframeinfo(client_frame)

                if sys.version_info >= (3, 11):
                    lineno = frameinfo.positions.lineno
                    end_lineno = frameinfo.positions.end_lineno
                    col_offset = frameinfo.positions.col_offset
                    end_col_offset = frameinfo.positions.end_col_offset
                else:
                    # Traceback in Python < 3.11 does not provide
                    # end line nor column information
                    lineno = frameinfo.lineno
                    end_lineno = lineno
                    col_offset = 0
                    end_col_offset = sys.maxsize

                self.recording.add(
                    client_frame.f_code.co_filename,
                    type_name,
                    member,
                    lineno,
                    end_lineno,
                    col_offset,
                    end_col_offset,
                )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides graphs for CommSystem objects."""
import collections.abc
import typing

import matplotlib

from ansys.stk.core.stkobjects import CommSystem
from ansys.stk.extensions.data_analysis.graphs.graph_helpers import line_chart


def carrier_to_noise_vs_time_line_chart(stk_object :CommSystem, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Graph the carrier-to-noise ratio and the carrier-to-noise-plus-interference ratio as a function of time.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\CommSystem\\Carrier to Noise vs Time.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.CommSystem
        The STK CommSystem object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Link Information').execute_elements(start_time, stop_time, step, ['C/N', 'Time', 'C/(N+I)']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Ratio', 'lines': [
            {'y_name':'c/n', 'label':'C/N', 'use_unit':True, 'unit_squared': None, 'dimension': 'Ratio'}]},
            {'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Ratio', 'lines': [
            {'y_name':'c/(n+i)', 'label':'C/(N+I)', 'use_unit':True, 'unit_squared': None, 'dimension': 'Ratio'}]}]
    return line_chart([df], root, ['c/n','c/(n+i)'], ['time'], axes, 'time', 'Time', 'Carrier to Noise vs Time', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)




# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides graphs for Scenario objects."""
import collections.abc
import typing

import matplotlib

from ansys.stk.core.stkobjects import Scenario
from ansys.stk.extensions.data_analysis.graphs.graph_helpers import line_chart


def greenwich_hour_angle_line_chart(stk_object :Scenario, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the Mean Greenwich Hour angle (Mean GHA) over time.

    Mean GHA is sidereal Greenwich Hour Angle (GHA) that does not include equation of the equinox.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\Scenario\\Greenwich Hour Angle.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.Scenario
        The STK Scenario object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('J2000 Angles').execute_elements(start_time, stop_time, step, ['Mean GHA', 'Time']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'mean gha', 'label':'Mean GHA', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]}]
    return line_chart([df], root, ['mean gha'], ['time'], axes, 'time', 'Time', 'Greenwich Hour Angle', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def polewanderx_line_chart(stk_object :Scenario, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the X component of pole wander over time.

    Pole wander describes the unmodeled part of the Earth's pole location.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\Scenario\\PoleWanderX.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.Scenario
        The STK Scenario object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('J2000 Angles').execute_elements(start_time, stop_time, step, ['Pole Wander X', 'Time']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'pole wander x', 'label':'Pole Wander X', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]}]
    return line_chart([df], root, ['pole wander x'], ['time'], axes, 'time', 'Time', 'PoleWanderX', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def polewandery_line_chart(stk_object :Scenario, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the Y component of pole wander over time.

    Pole wander describes the unmodeled part of the Earth's pole location.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\Scenario\\PoleWanderY.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.Scenario
        The STK Scenario object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('J2000 Angles').execute_elements(start_time, stop_time, step, ['Pole Wander Y', 'Time']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': True, 'label': 'Angle', 'lines': [
            {'y_name':'pole wander y', 'label':'Pole Wander Y', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]}]
    return line_chart([df], root, ['pole wander y'], ['time'], axes, 'time', 'Time', 'PoleWanderY', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def ut1_utc_line_chart(stk_object :Scenario, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the difference between the UT1 and UTC time scales over time.

    Civil time is tied to UTC while UT1 represents time using the actual rotation of the Earth, which varies due to gravitational effects of the Moon, the Sun, and other planets.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\Scenario\\UT1-UTC.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.Scenario
        The STK Scenario object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('J2000 Angles').execute_elements(start_time, stop_time, step, ['Time', 'UT1-UTC']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Time', 'lines': [
            {'y_name':'ut1-utc', 'label':'UT1-UTC', 'use_unit':True, 'unit_squared': None, 'dimension': 'Time'}]}]
    return line_chart([df], root, ['ut1-utc'], ['time'], axes, 'time', 'Time', 'UT1-UTC', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)


# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Contains specific exceptions that may be raised from the STK API."""


class STKInitializationError(RuntimeError):
    """Raised in STKDesktop and STKEngine when unable to initialize or attach to STK."""

class STKInvalidCastError(RuntimeError):
    """Raised when attempting to cast an object to an unsupported interface or class type."""

class STKRuntimeError(RuntimeError):
    """Raised when an STK method call fails."""

class STKAttributeError(AttributeError):
    """Raised when attempting to set an unrecognized attribute within the STK API."""

class STKEventsAPIError(SyntaxError):
    """Raised when attempting to assign to an STK Event rather than using operator += or -=."""

class STKPluginMethodNotImplementedError(SyntaxError):
    """Raised when a plugin method is called by STK that was not implemented by the user."""

class STKInvalidTimerError(RuntimeError):
    """Raised when attempting to use an engine timer that cannot be implemented."""

class STKColorError(RuntimeError):
    """Raised when a problem is encountered with color classes."""

class GrpcUtilitiesError(SyntaxError):
    """Raised when using gRPC utilities in an unsupported manner."""



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import gc
import typing

from ..utilities.exceptions import STKAttributeError, STKInitializationError, STKInvalidCastError

class InterfaceProxy(object):
    """Proxy class to isolate the call strategy to STK (COM, gRPC, etc)."""
    def __init__(self):
        pass

    def __eq__(self, other):
        """Check for equivalence of the underlying STK interface."""
        return False

    def __hash__(self):
        """Manage reference count."""
        return 0

    def __bool__(self):
        """Represent a valid interface."""
        return False

    def query_interface(self, intf_metadata:dict) -> "InterfaceProxy":
        """Return a new object with the requested guid."""
        return InterfaceProxy()

    def invoke(self, intf_metadata:dict, method_metadata:dict, *args):
        pass

    def get_property(self, intf_metadata:dict, method_metadata:dict):
        pass

    def set_property(self, intf_metadata:dict, method_metadata:dict, value):
        pass

class EnumeratorProxy(object):
    """Proxy class to isolate the call strategy for enumeration (COM, gRPC, etc)."""
    def __init__(self):
        pass

    def next(self) -> typing.Any:
        """Return the next item in the collection."""
        return None

    def reset(self):
        """Reset the enumeration of the collection."""

class SupportsDeleteCallback(object):
    """Execute callbacks on object deletion."""
    def __init__(self):
        self.__dict__["_del_callbacks"] = []

    def __del__(self):
        for callback in self._del_callbacks:
            callback()

    def _add_delete_callback(self, callback:typing.Callable) -> None:
        self._del_callbacks.append(callback)

class OutArg(object):
    pass

class GcDisabler(object):
    """Temporarily disable garbage collection."""
    def __init__(self):
        self._is_gc_enabled = False
    def __enter__(self):
        if gc.isenabled():
            self._is_gc_enabled = True
            gc.disable()
        return self
    def __exit__(self, type, value, tb):
        if self._is_gc_enabled:
            gc.enable()
        return False

def initialize_from_source_object(this, sourceObject, interfaceType):
    this.__dict__["_intf"] = InterfaceProxy()
    if sourceObject is not None and sourceObject._intf is not None:
        intf = sourceObject._intf.query_interface(interfaceType._metadata)
        if intf is not None:
            this._private_init(intf)
            del(intf)
        else:
            raise STKInvalidCastError(f"Failed to create {interfaceType.__name__} from source object.")

def get_interface_property(attrname, interfaceType):
    if attrname in interfaceType.__dict__ and type(interfaceType.__dict__[attrname]) == property:
        return interfaceType.__dict__[attrname]
    return None

def set_interface_attribute(this, attrname, value, interfaceType, baseType):
    if interfaceType._get_property(this, attrname) is not None:
        interfaceType._get_property(this, attrname).__set__(this, value)
    elif baseType is not None:
        baseType.__setattr__(this, attrname, value)
    else:
        raise STKAttributeError(f"{attrname} is not a recognized attribute in {interfaceType.__name__}.")

def set_class_attribute(this, attrname, value, classType, interfaceTypes):
    found_prop = None
    for interfaceType in interfaceTypes:
        found_prop_in_interface = interfaceType._get_property(this, attrname)
        if found_prop_in_interface is not None:
            found_prop = found_prop_in_interface
    if found_prop is not None:
        found_prop.__set__(this, value)
    else:
        raise STKAttributeError(f"{attrname} is not a recognized attribute in {classType.__name__}.")

def _unquoted(s:str) -> str:
    if s is not None and len(s) > 0:
        if s[0] == "\"":
            s = s[1:]
        if len(s) > 0 and s[-1] == "\"":
            s = s[:-1]
    return s

def read_registry_key(key, value=None, root=None, silent_exception=False):
    try:
        import winreg
        if root is None:
            root = winreg.HKEY_CLASSES_ROOT
        with winreg.OpenKey(root, key) as hkey:
            (val, typ) = winreg.QueryValueEx(hkey, value)
        return _unquoted(val)
    except Exception as e:
        if not silent_exception:
            raise STKInitializationError(f"Error Reading Registry for {key}: {e}")
        return None

def winreg_stk_binary_dir():
    try:
        import winreg
        return _unquoted(read_registry_key(f"SOFTWARE\\AGI\\STK\\12.0", root=winreg.HKEY_LOCAL_MACHINE, value="STKBinaryFolder"))
    except Exception as e:
        return None


# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Starts STK Engine and provides access to the Object Model root."""

__all__ = ["STKEngine", "STKEngineApplication", "STKEngineTimerType"]

import atexit
from ctypes import byref
from enum import IntEnum
import os

if os.name != "nt":
    from ctypes import CFUNCTYPE, cdll
    from ctypes.util import find_library

    from ..internal.timerutil import NullTimer, SigAlarmTimer, SigRtTimer, TclTimer
else:
    from ..internal.timerutil import NullTimer

from ..internal.comutil import (
    CLSCTX_INPROC_SERVER,
    GUID,
    CoInitializeManager,
    IUnknown,
    ObjectLifetimeManager,
    OLE32Lib,
    Succeeded,
)
from ..internal.eventutil import EventSubscriptionManager
from ..internal.stkxinitialization import STKXInitialize
from ..stkobjects import STKObjectModelContext, STKObjectRoot
from ..stkx import STKXApplication
from ..utilities.exceptions import GrpcUtilitiesError, STKInitializationError, STKRuntimeError
from ..utilities.grpcutilities import GrpcCallBatcher


class STKEngineTimerType(IntEnum):
    """
    Specify the timer implementation to use.

    Timers are needed for events on Linux applications.
    May be overridden by specifying environment variable STK_PYTHONAPI_TIMERTYPE.
    """

    DISABLE_TIMERS     = 1
    """Disable timers. This option is always selected on Windows applications."""
    TKINTER_MAIN_LOOP   = 2
    """Tkinter TCL timer dependent on the tkinter main loop. Default for tkinter applications."""
    INTERACTIVE_PYTHON = 3
    """Tkinter TCL timer dependent on the interactive Python environment. Default for interactive Python applications."""
    SIG_ALARM          = 4
    """Use the standard signal SIGALRM for timer notifications. Default when not using tkinter or interactice Python."""
    SIG_RT             = 5
    """Use a real-time signal for timer notifications. Default signal is SIGRTMIN. May be overridden by specifying environment variable STK_PYHONAPI_TIMERTYPE5_SIGRTMIN_OFFSET."""

class STKEngineApplication(STKXApplication):
    """
    Interact with STK Engine.

    Use STKEngine.StartApplication() to obtain an initialized STKEngineApplication object.
    """

    def __init__(self):
        """Construct an object of type STKEngineApplication."""
        STKXApplication.__init__(self)
        self.__dict__["_initialized"] = False
        self.__dict__["_grpc_exceptions"] = True

    def _private_init(self, unknown:IUnknown, no_graphics):
        STKXApplication._private_init(self, unknown)
        if os.name!="nt":
            self._stkx_intialize()
        self._stkx_intialize_timer(no_graphics)
        self.__dict__["_initialized"] = True

    def __del__(self):
        """Destruct the STKEngineApplication object after all references to the object are deleted."""
        self.shutdown()

    def _stkx_intialize(self):
        clsid_agstkxinitialize = GUID()
        if Succeeded(OLE32Lib.CLSIDFromString("{3B85901D-FC82-4733-97E6-5BB25CE69379}", clsid_agstkxinitialize)):
            iid_iunknown = GUID(IUnknown._guid)
            stkxinit_unk = IUnknown()
            if Succeeded(OLE32Lib.CoCreateInstance(byref(clsid_agstkxinitialize), None, CLSCTX_INPROC_SERVER, byref(iid_iunknown), byref(stkxinit_unk.p))):
                stkxinit_unk.take_ownership()
                stkxinit = STKXInitialize()
                stkxinit._private_init(stkxinit_unk)
                install_dir = os.getenv("STK_INSTALL_DIR")
                if install_dir is None:
                    raise STKInitializationError("Please set a valid STK_INSTALL_DIR environment variable.")
                config_dir = os.getenv("STK_CONFIG_DIR")
                if config_dir is None:
                    raise STKInitializationError("Please set a valid STK_CONFIG_DIR environment variable.")
                stkxinit.initialize_data(install_dir, config_dir)

    @staticmethod
    def _get_signo(sigrtmin_offset):
        if os.name=="nt":
            return None
        libc = cdll.LoadLibrary(find_library("c"))
        __libc_current_sigrtmin = CFUNCTYPE(c_int)(("__libc_current_sigrtmin", libc))
        return __libc_current_sigrtmin() + sigrtmin_offset

    def _set_timer_type_from_env(self):
        timer_type = int(os.getenv("STK_PYTHONAPI_TIMERTYPE", "4"))
        if os.name=="nt" or timer_type == STKEngineTimerType.DISABLE_TIMERS:
            self.__dict__["_timer_impl"] = NullTimer()
        elif os.name != "nt":
            if timer_type == STKEngineTimerType.TKINTER_MAIN_LOOP or timer_type == STKEngineTimerType.INTERACTIVE_PYTHON:
                self.__dict__["_timer_impl"] = TclTimer()
            elif timer_type == STKEngineTimerType.SIG_ALARM:
                self.__dict__["_timer_impl"] = SigAlarmTimer()
            elif timer_type == STKEngineTimerType.SIG_RT:
                sigrtmin_offset = int(os.getenv("STK_PYTHONAPI_TIMERTYPE5_SIGRTMIN_OFFSET", "0"))
                signo = STKEngineApplication._get_signo(sigrtmin_offset)
                self.__dict__["_timer_impl"] = SigRtTimer(signo)

    def _user_override_timer_type(self) -> bool:
        return ("STK_PYTHONAPI_TIMERTYPE" in os.environ)

    def _stkx_intialize_timer(self, no_graphics):
        if os.name=="nt":
            #Timers are not implemented on Windows, use a placeholder.
            self.__dict__["_timer_impl"] = NullTimer()
        elif no_graphics:
            self._set_timer_type_from_env()
        else:
            #default to Tkinter mainloop in graphics applications, but allow the user to override
            #e.g. if controls are not being used, the SIG_ALARM timer might be desired.
            if self._user_override_timer_type():
                self._set_timer_type_from_env()
            else:
                self.__dict__["_timer_impl"] = TclTimer()

    def new_object_root(self) -> STKObjectRoot:
        """Create a new object model root for the STK Engine application."""
        if not self.__dict__["_initialized"]:
            raise STKRuntimeError("STKEngineApplication has not been properly initialized.  Use StartApplication() to obtain the STKEngineApplication object.")
        clsid_agstkobjectroot = GUID()
        if Succeeded(OLE32Lib.CLSIDFromString("{96C1CE4E-C61D-4657-99CB-8581E12693FE}", clsid_agstkobjectroot)):
            iid_iunknown = GUID(IUnknown._guid)
            root_unk = IUnknown()
            OLE32Lib.CoCreateInstance(byref(clsid_agstkobjectroot), None, CLSCTX_INPROC_SERVER, byref(iid_iunknown), byref(root_unk.p))
            root_unk.take_ownership()
            root = STKObjectRoot()
            root._private_init(root_unk)
            return root

    def new_object_model_context(self) -> STKObjectModelContext:
        """Create a new object model context for the STK Engine application."""
        if not self.__dict__['_initialized']:
            raise STKRuntimeError('STKEngineApplication has not been properly initialized.  Use StartApplication() to obtain the STKEngineApplication object.')
        clsid_agstkobjectmodelcontext = GUID()
        if Succeeded(OLE32Lib.CLSIDFromString('{7A12879C-5018-4433-8415-5DB250AFBAF9}', clsid_agstkobjectmodelcontext)):
            iid_iunknown = GUID(IUnknown._guid)
            context_unk = IUnknown()
            OLE32Lib.CoCreateInstance(byref(clsid_agstkobjectmodelcontext), None, CLSCTX_INPROC_SERVER, byref(iid_iunknown), byref(context_unk.p))
            context_unk.take_ownership()
            context = STKObjectModelContext()
            context._private_init(context_unk)
            return context

    def set_grpc_options(self, options:dict) -> None:
        """
        Grpc is not available with STK Engine. Provided for parity with STK Runtime and Desktop.

        Available options include:
        { "raise exceptions with STK Engine" : bool }. Set to false to suppress exceptions when
        using SetGrpcOptions and NewGrpcCallBatcher with STK Engine.
        """
        if "raise exceptions with STK Engine" in options:
            self.__dict__["_grpc_exceptions"] = options["raise exceptions with STK Engine"]
        if self._grpc_exceptions:
            raise GrpcUtilitiesError("gRPC is not available with STK Engine. Disable this exception with SetGrpcOptions({\"raise exceptions with STK Engine\" : False}).")

    def new_grpc_call_batcher(self, max_batch:int=None, disable_batching:bool=True) -> GrpcCallBatcher:
        """Grpc is not available with STK Engine. Provided for parity with STK Runtime and Desktop."""
        if self._grpc_exceptions:
            raise GrpcUtilitiesError("gRPC is not available with STK Engine. Disable this exception with SetGrpcOptions({\"raise exceptions with STK Engine\" : False}).")
        return GrpcCallBatcher(disable_batching=True)

    def shutdown(self) -> None:
        """Shut down the STK Engine application."""
        if self._initialized:
            EventSubscriptionManager.unsubscribe_all()
            self._timer_impl.terminate()
            ObjectLifetimeManager.release_all(releaseApplication=False)
            self.terminate()
            ObjectLifetimeManager.release_all(releaseApplication=True)
            CoInitializeManager.uninitialize()
            self.__dict__["_initialized"] = False


class STKEngine(object):
    """Initialize and manage the STK Engine application."""

    _is_engine_running = False

    @staticmethod
    def _init_x11(no_graphics):
        if no_graphics or os.name=="nt":
            return
        try:
            x11lib = cdll.LoadLibrary(find_library("X11"))
            xinit_threads = CFUNCTYPE(None)(("XInitThreads", x11lib))
            xinit_threads()
        except OSError:
            raise STKRuntimeError("Failed attempting to run graphics mode without X11.")

    @staticmethod
    def start_application(no_graphics:bool=True) -> STKEngineApplication:
        """
        Initialize STK Engine in-process and return the instance.

        Must only be used once per Python process.
        """
        if STKEngine._is_engine_running:
            raise STKRuntimeError("Only one STKEngine instance is allowed per Python process.")
        CoInitializeManager.initialize()
        clsid_agstkxapplication = GUID()
        if Succeeded(OLE32Lib.CLSIDFromString("{062AB565-B121-45B5-A9A9-B412CEFAB6A9}", clsid_agstkxapplication)):
            unknown = IUnknown()
            iid_iunknown = GUID(IUnknown._guid)
            if Succeeded(OLE32Lib.CoCreateInstance(byref(clsid_agstkxapplication), None, CLSCTX_INPROC_SERVER, byref(iid_iunknown), byref(unknown.p))):
                unknown.take_ownership(isApplication=True)
                STKEngine._is_engine_running = True
                STKEngine._init_x11(no_graphics)
                engine = STKEngineApplication()
                engine._private_init(unknown, no_graphics)
                engine.no_graphics = no_graphics
                atexit.register(engine.shutdown)
                return engine
        raise STKInitializationError("Failed to create STK Engine application.  Check for successful install and registration.")


# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides graphs for CoverageDefinition objects."""
import collections.abc
import typing

import matplotlib

from ansys.stk.core.stkobjects import CoverageDefinition
from ansys.stk.extensions.data_analysis.graphs._common_graphs import (
    _coverage_by_latitude_xy_graph,
    _gap_duration_xy_graph,
)
from ansys.stk.extensions.data_analysis.graphs.graph_helpers import interval_plot, line_chart


def coverage_by_latitude_line_chart(stk_object : CoverageDefinition, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the percent time covered vs latitude.

    A point is considered to be covered if it has access to one or more of the assigned assets. The reported values for each latitude are the average value for all grid points at that latitude.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\CoverageDefinition\\Coverage By Latitude.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.CoverageDefinition
        The STK CoverageDefinition object.
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _coverage_by_latitude_xy_graph(stk_object, colormap, time_unit_abbreviation, formatter)

def gap_duration_line_chart(stk_object : CoverageDefinition, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the cumulative distribution of the access duration gaps of all grid points.

    For each grid point, access intervals to each assigned asset are combined to determine the time intervals over which at least one asset has access to the grid point. The durations of the gaps between these intervals, for all grid points, are then sorted from smallest to largest and the percentages are then plotted.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\CoverageDefinition\\Gap Duration.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.CoverageDefinition
        The STK CoverageDefinition object.
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _gap_duration_xy_graph(stk_object, colormap, time_unit_abbreviation, formatter)

def gi_point_coverage_interval_graph(stk_object :CoverageDefinition, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create an interval graph of the access intervals for the point currently selected by the grid inspector.

    The intervals represent the union of times that the grid point has access to any of the assets.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\CoverageDefinition\\GI Point Coverage.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.CoverageDefinition
        The STK CoverageDefinition object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Selected Point Coverage').execute_elements(start_time, stop_time, ['Access Start', 'Access End']).data_sets.to_pandas_dataframe()
    elements=[(('access start', 'None'),('access end', 'None'))]
    return interval_plot([df], root, elements, [], ['access start','access end'], 'Time', 'Point Coverage Intervals', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def gi_point_prob_of_coverage_line_chart(stk_object : CoverageDefinition, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the probability of coverage being achieved for the point selected in the grid inspector, as a function of the time past a request for coverage.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\CoverageDefinition\\GI Point Prob Of Coverage.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.CoverageDefinition
        The STK CoverageDefinition object.
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    df = stk_object.data_providers.item('Selected Point Probability of Coverage').execute_elements(['Time From Request', 'Probability of Collection']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : False, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Value', 'lines': [
            {'y_name':'probability of collection', 'label':'Probability of Collection', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'}]}]
    return line_chart([df], root, ['probability of collection', 'time from request'], [], axes, 'time from request','Time From Request', 'Probability of Coverage', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def gi_region_coverage_line_chart(stk_object :CoverageDefinition, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the percentage of coverage over time of the region selected by the grid inspector.

    A region is considered to be covered if at least one point within the region has access to one or more of the assigned assets.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\CoverageDefinition\\GI Region Coverage.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.CoverageDefinition
        The STK CoverageDefinition object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Selected Region Coverage').execute_elements(start_time, stop_time, step, ['Time', 'Percent Coverage']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : False, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Percent', 'lines': [
            {'y_name':'percent coverage', 'label':'% Coverage', 'use_unit':False, 'unit_squared': None, 'dimension': 'Percent'}]}]
    return line_chart([df], root, ['percent coverage'], ['time'], axes, 'time', 'Time', 'Percentage of Region Covered', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def gi_region_full_coverage_interval_graph(stk_object :CoverageDefinition, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create an interval graph of the intervals of time when the region selected by the grid inspector is completely covered.

    The region is considered to be completely covered when all points within the region are covered. A point is covered when it has access to some asset.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\CoverageDefinition\\GI Region Full Coverage.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.CoverageDefinition
        The STK CoverageDefinition object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Selected Region Full Coverage').execute_elements(start_time, stop_time, ['Coverage Start', 'Coverage End']).data_sets.to_pandas_dataframe()
    elements=[(('coverage start', 'None'),('coverage end', 'None'))]
    return interval_plot([df], root, elements, [], ['coverage start','coverage end'], 'Time', 'Intervals of Full Regional Coverage', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def gi_region_time_to_cover_line_chart(stk_object :CoverageDefinition, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the amount of wait time required, starting from the reported time, before complete coverage of the region selected in the grid inspector occurs.

    The average wait time, compute as the mean of samples, is also plotted. A region is considered to be completely covered if all points within the region have had access to at least one of the assigned assets.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\CoverageDefinition\\GI Region Time to Cover.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.CoverageDefinition
        The STK CoverageDefinition object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Selected Region Time To Cover').execute_elements(start_time, stop_time, step, ['Wait for Total Cov', 'Time', 'Average Sampled Wait']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Time', 'lines': [
            {'y_name':'wait for total cov', 'label':'Wait for Total Cov', 'use_unit':True, 'unit_squared': None, 'dimension': 'Time'},
            {'y_name':'average sampled wait', 'label':'Average Sampled Wait', 'use_unit':True, 'unit_squared': None, 'dimension': 'Time'}]}]
    return line_chart([df], root, ['wait for total cov','average sampled wait'], ['time'], axes, 'time', 'Time', 'GI Region Time to Cover', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def percent_coverage_line_chart(stk_object :CoverageDefinition, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the percentage of the area of the total coverage grid which is covered at the reported time.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\CoverageDefinition\\Percent Coverage.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.CoverageDefinition
        The STK CoverageDefinition object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Percent Coverage').execute_elements(start_time, stop_time, step, ['Percent Accum Coverage', 'Time', 'Percent Coverage']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : False, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Percent', 'lines': [
            {'y_name':'percent coverage', 'label':'% Coverage', 'use_unit':False, 'unit_squared': None, 'dimension': 'Percent'},
            {'y_name':'percent accum coverage', 'label':'% Accum Coverage', 'use_unit':False, 'unit_squared': None, 'dimension': 'Percent'}]}]
    return line_chart([df], root, ['percent coverage','percent accum coverage'], ['time'], axes, 'time', 'Time', 'Current and Accumulated Coverage', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)




# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Object Model components to support the MATLAB strategies found in the Basic Maneuver Procedure.

MATLAB strategies allow the users to utilize MATLAB
Aerospace Toolbox functions to define the force modeling of the aircraft.
"""

__all__ = ["BasicManeuverMATLABFactory", "StrategyMATLAB3DGuidance", "StrategyMATLABFull3D", "StrategyMATLABNavigation",
"StrategyMATLABProfile"]

from ctypes import POINTER
import typing

from ...internal import coclassutil as agcls, comutil as agcom, marshall as agmarshall
from ...internal.apiutil import (
    InterfaceProxy,
    OutArg,
    SupportsDeleteCallback,
    get_interface_property,
    initialize_from_source_object,
    set_class_attribute,
)
from ...internal.comutil import IUnknown
from ...stkobjects.aviator import ClosureMode, IAutomationStrategyFactory, IBasicManeuverStrategy
from ...utilities.exceptions import STKRuntimeError


def _raise_uninitialized_error(*args):
    raise STKRuntimeError("Valid STK object model classes are returned from STK methods and should not be created independently.")


class StrategyMATLABNavigation(IBasicManeuverStrategy, SupportsDeleteCallback):
    """Interface used to access options for a MATLAB - Horizontal Plane Strategy of a Basic Maneuver Procedure."""

    _num_methods = 7
    _vtable_offset = IUnknown._vtable_offset + IUnknown._num_methods
    _get_function_name_method_offset = 1
    _set_function_name_method_offset = 2
    _is_function_path_valid_method_offset = 3
    _get_check_for_errors_method_offset = 4
    _set_check_for_errors_method_offset = 5
    _get_display_output_method_offset = 6
    _set_display_output_method_offset = 7
    _metadata = {
        "iid_data" : (5227863432379747556, 7965596508726907792),
        "vtable_reference" : IUnknown._vtable_offset + IUnknown._num_methods - 1,
    }
    _property_names = {}
    def _get_property(self, attrname):
        return get_interface_property(attrname, StrategyMATLABNavigation)

    _get_function_name_metadata = { "offset" : _get_function_name_method_offset,
            "arg_types" : (POINTER(agcom.BSTR),),
            "marshallers" : (agmarshall.BStrArg,) }
    @property
    def function_name(self) -> str:
        """Get or set the name of the MATLAB function."""
        return self._intf.get_property(StrategyMATLABNavigation._metadata, StrategyMATLABNavigation._get_function_name_metadata)

    _set_function_name_metadata = { "offset" : _set_function_name_method_offset,
            "arg_types" : (agcom.BSTR,),
            "marshallers" : (agmarshall.BStrArg,) }
    @function_name.setter
    def function_name(self, value:str) -> None:
        """Get or set the name of the MATLAB function."""
        return self._intf.set_property(StrategyMATLABNavigation._metadata, StrategyMATLABNavigation._set_function_name_metadata, value)

    _is_function_path_valid_metadata = { "offset" : _is_function_path_valid_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    def is_function_path_valid(self) -> bool:
        """Check if the MATLAB function path is valid."""
        return self._intf.invoke(StrategyMATLABNavigation._metadata, StrategyMATLABNavigation._is_function_path_valid_metadata, OutArg())

    _get_check_for_errors_metadata = { "offset" : _get_check_for_errors_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @property
    def check_for_errors(self) -> bool:
        """Get or set the option to check the function for errors."""
        return self._intf.get_property(StrategyMATLABNavigation._metadata, StrategyMATLABNavigation._get_check_for_errors_metadata)

    _set_check_for_errors_metadata = { "offset" : _set_check_for_errors_method_offset,
            "arg_types" : (agcom.VARIANT_BOOL,),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @check_for_errors.setter
    def check_for_errors(self, value:bool) -> None:
        """Get or set the option to check the function for errors."""
        return self._intf.set_property(StrategyMATLABNavigation._metadata, StrategyMATLABNavigation._set_check_for_errors_metadata, value)

    _get_display_output_metadata = { "offset" : _get_display_output_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @property
    def display_output(self) -> bool:
        """Get or set the option to display the output from the MATLAB function."""
        return self._intf.get_property(StrategyMATLABNavigation._metadata, StrategyMATLABNavigation._get_display_output_metadata)

    _set_display_output_metadata = { "offset" : _set_display_output_method_offset,
            "arg_types" : (agcom.VARIANT_BOOL,),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @display_output.setter
    def display_output(self, value:bool) -> None:
        """Get or set the option to display the output from the MATLAB function."""
        return self._intf.set_property(StrategyMATLABNavigation._metadata, StrategyMATLABNavigation._set_display_output_metadata, value)

    _property_names[function_name] = "function_name"
    _property_names[check_for_errors] = "check_for_errors"
    _property_names[display_output] = "display_output"

    def __init__(self, source_object=None):
        """Construct an object of type StrategyMATLABNavigation."""
        SupportsDeleteCallback.__init__(self)
        initialize_from_source_object(self, source_object, StrategyMATLABNavigation)
        IBasicManeuverStrategy.__init__(self, source_object)
    def _private_init(self, intf:InterfaceProxy):
        self.__dict__["_intf"] = intf
        IBasicManeuverStrategy._private_init(self, intf)
    def __eq__(self, other):
        """Check equality of the underlying STK references."""
        return agcls.compare_com_objects(self, other)
    def __setattr__(self, attrname, value):
        """Attempt to assign an attribute."""
        set_class_attribute(self, attrname, value, StrategyMATLABNavigation, [StrategyMATLABNavigation, IBasicManeuverStrategy])

agcls.AgClassCatalog.add_catalog_entry((4922865625757889154, 4997589758508718220), StrategyMATLABNavigation)
agcls.AgTypeNameMap["StrategyMATLABNavigation"] = StrategyMATLABNavigation

class StrategyMATLABProfile(IBasicManeuverStrategy, SupportsDeleteCallback):
    """Interface used to access options for a MATLAB - Vertical Plane Strategy of a Basic Maneuver Procedure."""

    _num_methods = 7
    _vtable_offset = IUnknown._vtable_offset + IUnknown._num_methods
    _get_function_name_method_offset = 1
    _set_function_name_method_offset = 2
    _is_function_path_valid_method_offset = 3
    _get_check_for_errors_method_offset = 4
    _set_check_for_errors_method_offset = 5
    _get_display_output_method_offset = 6
    _set_display_output_method_offset = 7
    _metadata = {
        "iid_data" : (5762811461223097488, 815303183833033109),
        "vtable_reference" : IUnknown._vtable_offset + IUnknown._num_methods - 1,
    }
    _property_names = {}
    def _get_property(self, attrname):
        return get_interface_property(attrname, StrategyMATLABProfile)

    _get_function_name_metadata = { "offset" : _get_function_name_method_offset,
            "arg_types" : (POINTER(agcom.BSTR),),
            "marshallers" : (agmarshall.BStrArg,) }
    @property
    def function_name(self) -> str:
        """Get or set the name of the MATLAB function."""
        return self._intf.get_property(StrategyMATLABProfile._metadata, StrategyMATLABProfile._get_function_name_metadata)

    _set_function_name_metadata = { "offset" : _set_function_name_method_offset,
            "arg_types" : (agcom.BSTR,),
            "marshallers" : (agmarshall.BStrArg,) }
    @function_name.setter
    def function_name(self, value:str) -> None:
        """Get or set the name of the MATLAB function."""
        return self._intf.set_property(StrategyMATLABProfile._metadata, StrategyMATLABProfile._set_function_name_metadata, value)

    _is_function_path_valid_metadata = { "offset" : _is_function_path_valid_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    def is_function_path_valid(self) -> bool:
        """Check if the MATLAB function path is valid."""
        return self._intf.invoke(StrategyMATLABProfile._metadata, StrategyMATLABProfile._is_function_path_valid_metadata, OutArg())

    _get_check_for_errors_metadata = { "offset" : _get_check_for_errors_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @property
    def check_for_errors(self) -> bool:
        """Get or set the option to check the function for errors."""
        return self._intf.get_property(StrategyMATLABProfile._metadata, StrategyMATLABProfile._get_check_for_errors_metadata)

    _set_check_for_errors_metadata = { "offset" : _set_check_for_errors_method_offset,
            "arg_types" : (agcom.VARIANT_BOOL,),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @check_for_errors.setter
    def check_for_errors(self, value:bool) -> None:
        """Get or set the option to check the function for errors."""
        return self._intf.set_property(StrategyMATLABProfile._metadata, StrategyMATLABProfile._set_check_for_errors_metadata, value)

    _get_display_output_metadata = { "offset" : _get_display_output_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @property
    def display_output(self) -> bool:
        """Get or set the option to display the output from the MATLAB function."""
        return self._intf.get_property(StrategyMATLABProfile._metadata, StrategyMATLABProfile._get_display_output_metadata)

    _set_display_output_metadata = { "offset" : _set_display_output_method_offset,
            "arg_types" : (agcom.VARIANT_BOOL,),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @display_output.setter
    def display_output(self, value:bool) -> None:
        """Get or set the option to display the output from the MATLAB function."""
        return self._intf.set_property(StrategyMATLABProfile._metadata, StrategyMATLABProfile._set_display_output_metadata, value)

    _property_names[function_name] = "function_name"
    _property_names[check_for_errors] = "check_for_errors"
    _property_names[display_output] = "display_output"

    def __init__(self, source_object=None):
        """Construct an object of type StrategyMATLABProfile."""
        SupportsDeleteCallback.__init__(self)
        initialize_from_source_object(self, source_object, StrategyMATLABProfile)
        IBasicManeuverStrategy.__init__(self, source_object)
    def _private_init(self, intf:InterfaceProxy):
        self.__dict__["_intf"] = intf
        IBasicManeuverStrategy._private_init(self, intf)
    def __eq__(self, other):
        """Check equality of the underlying STK references."""
        return agcls.compare_com_objects(self, other)
    def __setattr__(self, attrname, value):
        """Attempt to assign an attribute."""
        set_class_attribute(self, attrname, value, StrategyMATLABProfile, [StrategyMATLABProfile, IBasicManeuverStrategy])

agcls.AgClassCatalog.add_catalog_entry((5431676618088683906, 7176732423655266234), StrategyMATLABProfile)
agcls.AgTypeNameMap["StrategyMATLABProfile"] = StrategyMATLABProfile

class StrategyMATLABFull3D(IBasicManeuverStrategy, SupportsDeleteCallback):
    """Interface used to access options for a MATLAB - Full 3D Strategy of a Basic Maneuver Procedure."""

    _num_methods = 7
    _vtable_offset = IUnknown._vtable_offset + IUnknown._num_methods
    _get_function_name_method_offset = 1
    _set_function_name_method_offset = 2
    _is_function_path_valid_method_offset = 3
    _get_check_for_errors_method_offset = 4
    _set_check_for_errors_method_offset = 5
    _get_display_output_method_offset = 6
    _set_display_output_method_offset = 7
    _metadata = {
        "iid_data" : (4991766284198429486, 11343135184924199837),
        "vtable_reference" : IUnknown._vtable_offset + IUnknown._num_methods - 1,
    }
    _property_names = {}
    def _get_property(self, attrname):
        return get_interface_property(attrname, StrategyMATLABFull3D)

    _get_function_name_metadata = { "offset" : _get_function_name_method_offset,
            "arg_types" : (POINTER(agcom.BSTR),),
            "marshallers" : (agmarshall.BStrArg,) }
    @property
    def function_name(self) -> str:
        """Get or set the name of the MATLAB function."""
        return self._intf.get_property(StrategyMATLABFull3D._metadata, StrategyMATLABFull3D._get_function_name_metadata)

    _set_function_name_metadata = { "offset" : _set_function_name_method_offset,
            "arg_types" : (agcom.BSTR,),
            "marshallers" : (agmarshall.BStrArg,) }
    @function_name.setter
    def function_name(self, value:str) -> None:
        """Get or set the name of the MATLAB function."""
        return self._intf.set_property(StrategyMATLABFull3D._metadata, StrategyMATLABFull3D._set_function_name_metadata, value)

    _is_function_path_valid_metadata = { "offset" : _is_function_path_valid_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    def is_function_path_valid(self) -> bool:
        """Check if the MATLAB function path is valid."""
        return self._intf.invoke(StrategyMATLABFull3D._metadata, StrategyMATLABFull3D._is_function_path_valid_metadata, OutArg())

    _get_check_for_errors_metadata = { "offset" : _get_check_for_errors_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @property
    def check_for_errors(self) -> bool:
        """Get or set the option to check the function for errors."""
        return self._intf.get_property(StrategyMATLABFull3D._metadata, StrategyMATLABFull3D._get_check_for_errors_metadata)

    _set_check_for_errors_metadata = { "offset" : _set_check_for_errors_method_offset,
            "arg_types" : (agcom.VARIANT_BOOL,),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @check_for_errors.setter
    def check_for_errors(self, value:bool) -> None:
        """Get or set the option to check the function for errors."""
        return self._intf.set_property(StrategyMATLABFull3D._metadata, StrategyMATLABFull3D._set_check_for_errors_metadata, value)

    _get_display_output_metadata = { "offset" : _get_display_output_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @property
    def display_output(self) -> bool:
        """Get or set the option to display the output from the MATLAB function."""
        return self._intf.get_property(StrategyMATLABFull3D._metadata, StrategyMATLABFull3D._get_display_output_metadata)

    _set_display_output_metadata = { "offset" : _set_display_output_method_offset,
            "arg_types" : (agcom.VARIANT_BOOL,),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @display_output.setter
    def display_output(self, value:bool) -> None:
        """Get or set the option to display the output from the MATLAB function."""
        return self._intf.set_property(StrategyMATLABFull3D._metadata, StrategyMATLABFull3D._set_display_output_metadata, value)

    _property_names[function_name] = "function_name"
    _property_names[check_for_errors] = "check_for_errors"
    _property_names[display_output] = "display_output"

    def __init__(self, source_object=None):
        """Construct an object of type StrategyMATLABFull3D."""
        SupportsDeleteCallback.__init__(self)
        initialize_from_source_object(self, source_object, StrategyMATLABFull3D)
        IBasicManeuverStrategy.__init__(self, source_object)
    def _private_init(self, intf:InterfaceProxy):
        self.__dict__["_intf"] = intf
        IBasicManeuverStrategy._private_init(self, intf)
    def __eq__(self, other):
        """Check equality of the underlying STK references."""
        return agcls.compare_com_objects(self, other)
    def __setattr__(self, attrname, value):
        """Attempt to assign an attribute."""
        set_class_attribute(self, attrname, value, StrategyMATLABFull3D, [StrategyMATLABFull3D, IBasicManeuverStrategy])

agcls.AgClassCatalog.add_catalog_entry((5699885202707284005, 15299874748251640476), StrategyMATLABFull3D)
agcls.AgTypeNameMap["StrategyMATLABFull3D"] = StrategyMATLABFull3D

class StrategyMATLAB3DGuidance(IBasicManeuverStrategy, SupportsDeleteCallback):
    """Interface used to access options for a MATLAB - 3D Guidance Strategy of a Basic Maneuver Procedure."""

    _num_methods = 29
    _vtable_offset = IUnknown._vtable_offset + IUnknown._num_methods
    _get_target_name_method_offset = 1
    _set_target_name_method_offset = 2
    _get_valid_target_names_method_offset = 3
    _get_target_resolution_method_offset = 4
    _set_target_resolution_method_offset = 5
    _get_use_stop_time_to_go_method_offset = 6
    _get_stop_time_to_go_method_offset = 7
    _set_stop_time_to_go_method_offset = 8
    _get_use_stop_slant_range_method_offset = 9
    _get_stop_slant_range_method_offset = 10
    _set_stop_slant_range_method_offset = 11
    _get_function_name_method_offset = 12
    _set_function_name_method_offset = 13
    _is_function_path_valid_method_offset = 14
    _get_check_for_errors_method_offset = 15
    _set_check_for_errors_method_offset = 16
    _get_display_output_method_offset = 17
    _set_display_output_method_offset = 18
    _get_closure_mode_method_offset = 19
    _set_closure_mode_method_offset = 20
    _get_hobs_max_angle_method_offset = 21
    _set_hobs_max_angle_method_offset = 22
    _get_hobs_angle_tol_method_offset = 23
    _set_hobs_angle_tol_method_offset = 24
    _get_compute_tas_dot_method_offset = 25
    _set_compute_tas_dot_method_offset = 26
    _get_airspeed_options_method_offset = 27
    _get_position_velocity_strategies_method_offset = 28
    _cancel_target_position_velocity_method_offset = 29
    _metadata = {
        "iid_data" : (5207808632699689454, 11389872158525712815),
        "vtable_reference" : IUnknown._vtable_offset + IUnknown._num_methods - 1,
    }
    _property_names = {}
    def _get_property(self, attrname):
        return get_interface_property(attrname, StrategyMATLAB3DGuidance)

    _get_target_name_metadata = { "offset" : _get_target_name_method_offset,
            "arg_types" : (POINTER(agcom.BSTR),),
            "marshallers" : (agmarshall.BStrArg,) }
    @property
    def target_name(self) -> str:
        """Get or set the target name."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_target_name_metadata)

    _set_target_name_metadata = { "offset" : _set_target_name_method_offset,
            "arg_types" : (agcom.BSTR,),
            "marshallers" : (agmarshall.BStrArg,) }
    @target_name.setter
    def target_name(self, value:str) -> None:
        """Get or set the target name."""
        return self._intf.set_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._set_target_name_metadata, value)

    _get_valid_target_names_metadata = { "offset" : _get_valid_target_names_method_offset,
            "arg_types" : (POINTER(agcom.LPSAFEARRAY),),
            "marshallers" : (agmarshall.LPSafearrayArg,) }
    @property
    def valid_target_names(self) -> list:
        """Return the valid target names."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_valid_target_names_metadata)

    _get_target_resolution_metadata = { "offset" : _get_target_resolution_method_offset,
            "arg_types" : (POINTER(agcom.DOUBLE),),
            "marshallers" : (agmarshall.DoubleArg,) }
    @property
    def target_resolution(self) -> float:
        """Get or set the target position/velocity sampling resolution."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_target_resolution_metadata)

    _set_target_resolution_metadata = { "offset" : _set_target_resolution_method_offset,
            "arg_types" : (agcom.DOUBLE,),
            "marshallers" : (agmarshall.DoubleArg,) }
    @target_resolution.setter
    def target_resolution(self, value:float) -> None:
        """Get or set the target position/velocity sampling resolution."""
        return self._intf.set_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._set_target_resolution_metadata, value)

    _get_use_stop_time_to_go_metadata = { "offset" : _get_use_stop_time_to_go_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @property
    def use_stop_time_to_go(self) -> bool:
        """Get the option to specify a time to go stopping condition."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_use_stop_time_to_go_metadata)

    _get_stop_time_to_go_metadata = { "offset" : _get_stop_time_to_go_method_offset,
            "arg_types" : (POINTER(agcom.DOUBLE),),
            "marshallers" : (agmarshall.DoubleArg,) }
    @property
    def stop_time_to_go(self) -> float:
        """Get the stop time from the target at which the maneuver will stop."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_stop_time_to_go_metadata)

    _set_stop_time_to_go_metadata = { "offset" : _set_stop_time_to_go_method_offset,
            "arg_types" : (agcom.VARIANT_BOOL, agcom.DOUBLE,),
            "marshallers" : (agmarshall.VariantBoolArg, agmarshall.DoubleArg,) }
    def set_stop_time_to_go(self, enable:bool, time:float) -> None:
        """Set the option to use the stop time from target stopping condition and set the according value."""
        return self._intf.invoke(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._set_stop_time_to_go_metadata, enable, time)

    _get_use_stop_slant_range_metadata = { "offset" : _get_use_stop_slant_range_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @property
    def use_stop_slant_range(self) -> bool:
        """Get the option to specify a range from target stopping condition."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_use_stop_slant_range_metadata)

    _get_stop_slant_range_metadata = { "offset" : _get_stop_slant_range_method_offset,
            "arg_types" : (POINTER(agcom.DOUBLE),),
            "marshallers" : (agmarshall.DoubleArg,) }
    @property
    def stop_slant_range(self) -> float:
        """Get the range from the target at which the maneuver will stop."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_stop_slant_range_metadata)

    _set_stop_slant_range_metadata = { "offset" : _set_stop_slant_range_method_offset,
            "arg_types" : (agcom.VARIANT_BOOL, agcom.DOUBLE,),
            "marshallers" : (agmarshall.VariantBoolArg, agmarshall.DoubleArg,) }
    def set_stop_slant_range(self, enable:bool, range:float) -> None:
        """Set the option to use the stop slant range stopping condition and set the according value."""
        return self._intf.invoke(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._set_stop_slant_range_metadata, enable, range)

    _get_function_name_metadata = { "offset" : _get_function_name_method_offset,
            "arg_types" : (POINTER(agcom.BSTR),),
            "marshallers" : (agmarshall.BStrArg,) }
    @property
    def function_name(self) -> str:
        """Get or set the name of the MATLAB function."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_function_name_metadata)

    _set_function_name_metadata = { "offset" : _set_function_name_method_offset,
            "arg_types" : (agcom.BSTR,),
            "marshallers" : (agmarshall.BStrArg,) }
    @function_name.setter
    def function_name(self, value:str) -> None:
        """Get or set the name of the MATLAB function."""
        return self._intf.set_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._set_function_name_metadata, value)

    _is_function_path_valid_metadata = { "offset" : _is_function_path_valid_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    def is_function_path_valid(self) -> bool:
        """Check if the MATLAB function path is valid."""
        return self._intf.invoke(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._is_function_path_valid_metadata, OutArg())

    _get_check_for_errors_metadata = { "offset" : _get_check_for_errors_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @property
    def check_for_errors(self) -> bool:
        """Get or set the option to check the function for errors."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_check_for_errors_metadata)

    _set_check_for_errors_metadata = { "offset" : _set_check_for_errors_method_offset,
            "arg_types" : (agcom.VARIANT_BOOL,),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @check_for_errors.setter
    def check_for_errors(self, value:bool) -> None:
        """Get or set the option to check the function for errors."""
        return self._intf.set_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._set_check_for_errors_metadata, value)

    _get_display_output_metadata = { "offset" : _get_display_output_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @property
    def display_output(self) -> bool:
        """Get or set the option to display the output from the MATLAB function."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_display_output_metadata)

    _set_display_output_metadata = { "offset" : _set_display_output_method_offset,
            "arg_types" : (agcom.VARIANT_BOOL,),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @display_output.setter
    def display_output(self, value:bool) -> None:
        """Get or set the option to display the output from the MATLAB function."""
        return self._intf.set_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._set_display_output_metadata, value)

    _get_closure_mode_metadata = { "offset" : _get_closure_mode_method_offset,
            "arg_types" : (POINTER(agcom.LONG),),
            "marshallers" : (agmarshall.EnumArg(ClosureMode),) }
    @property
    def closure_mode(self) -> "ClosureMode":
        """Get or set the closure mode for the guidance strategy."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_closure_mode_metadata)

    _set_closure_mode_metadata = { "offset" : _set_closure_mode_method_offset,
            "arg_types" : (agcom.LONG,),
            "marshallers" : (agmarshall.EnumArg(ClosureMode),) }
    @closure_mode.setter
    def closure_mode(self, value:"ClosureMode") -> None:
        """Get or set the closure mode for the guidance strategy."""
        return self._intf.set_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._set_closure_mode_metadata, value)

    _get_hobs_max_angle_metadata = { "offset" : _get_hobs_max_angle_method_offset,
            "arg_types" : (POINTER(agcom.Variant),),
            "marshallers" : (agmarshall.VariantArg,) }
    @property
    def hobs_max_angle(self) -> typing.Any:
        """Get or set the closure high off boresight max angle."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_hobs_max_angle_metadata)

    _set_hobs_max_angle_metadata = { "offset" : _set_hobs_max_angle_method_offset,
            "arg_types" : (agcom.Variant,),
            "marshallers" : (agmarshall.VariantArg,) }
    @hobs_max_angle.setter
    def hobs_max_angle(self, value:typing.Any) -> None:
        """Get or set the closure high off boresight max angle."""
        return self._intf.set_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._set_hobs_max_angle_metadata, value)

    _get_hobs_angle_tol_metadata = { "offset" : _get_hobs_angle_tol_method_offset,
            "arg_types" : (POINTER(agcom.Variant),),
            "marshallers" : (agmarshall.VariantArg,) }
    @property
    def hobs_angle_tol(self) -> typing.Any:
        """Get or set the closure high off boresight angle tolerance."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_hobs_angle_tol_metadata)

    _set_hobs_angle_tol_metadata = { "offset" : _set_hobs_angle_tol_method_offset,
            "arg_types" : (agcom.Variant,),
            "marshallers" : (agmarshall.VariantArg,) }
    @hobs_angle_tol.setter
    def hobs_angle_tol(self, value:typing.Any) -> None:
        """Get or set the closure high off boresight angle tolerance."""
        return self._intf.set_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._set_hobs_angle_tol_metadata, value)

    _get_compute_tas_dot_metadata = { "offset" : _get_compute_tas_dot_method_offset,
            "arg_types" : (POINTER(agcom.VARIANT_BOOL),),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @property
    def compute_tas_dot(self) -> bool:
        """Get or set the option to allow MATLAB to compute the true airspeed for the aircraft."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_compute_tas_dot_metadata)

    _set_compute_tas_dot_metadata = { "offset" : _set_compute_tas_dot_method_offset,
            "arg_types" : (agcom.VARIANT_BOOL,),
            "marshallers" : (agmarshall.VariantBoolArg,) }
    @compute_tas_dot.setter
    def compute_tas_dot(self, value:bool) -> None:
        """Get or set the option to allow MATLAB to compute the true airspeed for the aircraft."""
        return self._intf.set_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._set_compute_tas_dot_metadata, value)

    _get_airspeed_options_metadata = { "offset" : _get_airspeed_options_method_offset,
            "arg_types" : (POINTER(agcom.PVOID),),
            "marshallers" : (agmarshall.InterfaceOutArg,) }
    @property
    def airspeed_options(self) -> "IBasicManeuverAirspeedOptions":
        """Get the airspeed options."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_airspeed_options_metadata)

    _get_position_velocity_strategies_metadata = { "offset" : _get_position_velocity_strategies_method_offset,
            "arg_types" : (POINTER(agcom.PVOID),),
            "marshallers" : (agmarshall.InterfaceOutArg,) }
    @property
    def position_velocity_strategies(self) -> "IBasicManeuverTargetPositionVelocity":
        """Get the position velocity strategies for MATLAB 3D Guidance."""
        return self._intf.get_property(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._get_position_velocity_strategies_metadata)

    _cancel_target_position_velocity_metadata = { "offset" : _cancel_target_position_velocity_method_offset,
            "arg_types" : (),
            "marshallers" : () }
    def cancel_target_position_velocity(self) -> None:
        """Cancel the position velocity strategies for MATLAB 3D Guidance."""
        return self._intf.invoke(StrategyMATLAB3DGuidance._metadata, StrategyMATLAB3DGuidance._cancel_target_position_velocity_metadata, )

    _property_names[target_name] = "target_name"
    _property_names[valid_target_names] = "valid_target_names"
    _property_names[target_resolution] = "target_resolution"
    _property_names[use_stop_time_to_go] = "use_stop_time_to_go"
    _property_names[stop_time_to_go] = "stop_time_to_go"
    _property_names[use_stop_slant_range] = "use_stop_slant_range"
    _property_names[stop_slant_range] = "stop_slant_range"
    _property_names[function_name] = "function_name"
    _property_names[check_for_errors] = "check_for_errors"
    _property_names[display_output] = "display_output"
    _property_names[closure_mode] = "closure_mode"
    _property_names[hobs_max_angle] = "hobs_max_angle"
    _property_names[hobs_angle_tol] = "hobs_angle_tol"
    _property_names[compute_tas_dot] = "compute_tas_dot"
    _property_names[airspeed_options] = "airspeed_options"
    _property_names[position_velocity_strategies] = "position_velocity_strategies"

    def __init__(self, source_object=None):
        """Construct an object of type StrategyMATLAB3DGuidance."""
        SupportsDeleteCallback.__init__(self)
        initialize_from_source_object(self, source_object, StrategyMATLAB3DGuidance)
        IBasicManeuverStrategy.__init__(self, source_object)
    def _private_init(self, intf:InterfaceProxy):
        self.__dict__["_intf"] = intf
        IBasicManeuverStrategy._private_init(self, intf)
    def __eq__(self, other):
        """Check equality of the underlying STK references."""
        return agcls.compare_com_objects(self, other)
    def __setattr__(self, attrname, value):
        """Attempt to assign an attribute."""
        set_class_attribute(self, attrname, value, StrategyMATLAB3DGuidance, [StrategyMATLAB3DGuidance, IBasicManeuverStrategy])

agcls.AgClassCatalog.add_catalog_entry((4932746689876178541, 12483900520759435676), StrategyMATLAB3DGuidance)
agcls.AgTypeNameMap["StrategyMATLAB3DGuidance"] = StrategyMATLAB3DGuidance

class BasicManeuverMATLABFactory(IAutomationStrategyFactory, SupportsDeleteCallback):
    """Class defining the factory to create the basic maneuver PropNav strategies."""
    def __init__(self, source_object=None):
        """Construct an object of type BasicManeuverMATLABFactory."""
        SupportsDeleteCallback.__init__(self)
        IAutomationStrategyFactory.__init__(self, source_object)
    def _private_init(self, intf:InterfaceProxy):
        self.__dict__["_intf"] = intf
        IAutomationStrategyFactory._private_init(self, intf)
    def __eq__(self, other):
        """Check equality of the underlying STK references."""
        return agcls.compare_com_objects(self, other)
    def __setattr__(self, attrname, value):
        """Attempt to assign an attribute."""
        set_class_attribute(self, attrname, value, BasicManeuverMATLABFactory, [IAutomationStrategyFactory])

agcls.AgClassCatalog.add_catalog_entry((5583954005185604195, 6409795893672302240), BasicManeuverMATLABFactory)
agcls.AgTypeNameMap["BasicManeuverMATLABFactory"] = BasicManeuverMATLABFactory